"""
Provides functions for interacting with MS data through ProteoWizard.
"""

from __future__ import absolute_import, division

import cgi
import logging
import os
import platform
import shutil
import subprocess
import sys
import tempfile

import requests
import pymzml


LOGGER = logging.getLogger("pycamv.proteowizard")


THIS_DIR = os.path.abspath(os.path.dirname(__file__))
PROTEOWIZARD_DIR = os.path.join(THIS_DIR, "ProteoWizard")

PROTEOWIZARD_VERSION = "3.0.10577"
PROTEOWIZARD_PATH = os.path.join(
    PROTEOWIZARD_DIR,
    "ProteoWizard {}".format(PROTEOWIZARD_VERSION),
)

PROTEOWIZARD_MSI_32_URLS = [
    (
        (
            "https://www.dropbox.com/s/we46o1hvkzlpel7/"
            "pwiz-setup-3.0.10577-x86.msi?dl=1"
        ),
        "GET",
        None,
    ),
    (
        "http://data.mallicklab.com/download.php",
        "POST",
        {"downloadtype": "bt36i"},
    ),
]

PROTEOWIZARD_MSI_64_URLS = [
    (
        (
            "https://www.dropbox.com/s/9rawc44u84jr9ip/"
            "pwiz-setup-3.0.10577-x86_64.msi?dl=1"
        ),
        "GET",
        None,
    ),
    (
        "http://data.mallicklab.com/download.php",
        "POST",
        {"downloadtype": "bt83i"},
    ),
]


def fetch_proteowizard(urls=None):
    LOGGER.debug("Proteowizard: {}".format(PROTEOWIZARD_PATH))

    if os.path.exists(PROTEOWIZARD_PATH):
        return

    LOGGER.info("ProteoWizard not installed, fetching now.")

    if platform.system() not in ["Windows"]:
        raise Exception("Proteowizard install not supported on your platform")

    if urls is None:
        if platform.architecture()[0] == "64bit":
            urls = PROTEOWIZARD_MSI_64_URLS
        else:
            urls = PROTEOWIZARD_MSI_32_URLS

    tmpdir = tempfile.mkdtemp()

    # Download the .msi file
    responses = []
    for url, req_type, post in urls:
        if req_type in ["GET"]:
            response = requests.get(url, stream=True)
        elif req_type in ["POST"]:
            response = requests.post(url, data=post, stream=True)
        else:
            raise Exception("Unknown request type: {}".format(req_type))

        if not response.ok:
            responses.append(response)
            continue

        out_path = os.path.join(
            tmpdir,
            cgi.parse_header(
                response.headers["Content-Disposition"]
            )[-1]["filename"]
        )

        with open(out_path, mode="wb") as f:
            for block in response.iter_content(1024):
                # print(block)
                # return
                f.write(block)

            break
    else:
        raise Exception("Unable to download file: {}".format(responses))

    # Extract the msi file's contents
    extract_path = os.path.join(tmpdir, "msi_extract")
    cmd = [
        "msiexec",
        "/a",
        out_path,
        "/qb",
        "TARGETDIR=\"{}\"".format(extract_path),
    ]
    subprocess.check_call(" ".join(cmd), shell=True)

    # Copy the msi file's contents to PROTEOWIZARD_DIR
    src = os.path.join(
        extract_path,
        "PFiles",
        "ProteoWizard",
    )

    shutil.rmtree(PROTEOWIZARD_DIR, ignore_errors=True)
    shutil.copytree(src, PROTEOWIZARD_DIR)
    shutil.rmtree(tmpdir)


def raw_to_mzml(raw_path, out_dir, scans=None, mz_window=None):
    """
    Covert a RAW file to .mzML using ProteoWizard.

    Parameters
    ----------
    raw_path : str
    out_dir : str
    scans : list of int, optional
    mz_window : list of int, optional

    Returns
    -------
    :class:`pymzml.run.Reader<run.Reader>`
    """
    fetch_proteowizard()

    ms_convert_path = os.path.join(PROTEOWIZARD_PATH, "msconvert.exe")

    # Create a config file,
    config, config_path = tempfile.mkstemp(suffix=".txt", text=True)

    with os.fdopen(config, "w+") as config:
        if scans:
            config.write(
                "filter=\"scanNumber {}\"\n".format(
                    " ".join(str(scan) for scan in scans)
                )
            )

        if mz_window:
            config.write(
                "filter=\"mzWindow [{},{}]\"\n".format(
                    mz_window[0], mz_window[1],
                )
            )

    # Run msconvert to convert raw file to mzML
    LOGGER.info("Converting \"{}\" to .mzML format.".format(raw_path))

    cmd = [
        ms_convert_path,
        raw_path,
        "-o", out_dir,
        "--mzML",
        "-c", config_path,
    ]

    out = subprocess.check_output(cmd)
    encoding = sys.stdout.encoding or "utf-8"
    LOGGER.debug(out.decode(encoding))

    os.remove(config_path)

    # Read the file into memory using pymzml
    basename = os.path.splitext(os.path.basename(raw_path))[0]
    out_path = os.path.join(out_dir, "{}.mzML".format(basename))
    data = pymzml.run.Reader(
        out_path,
        extraAccessions=[
            ("MS:1000827", ["value"]),  # isolation window target m/z
            ("MS:1000828", ["value"]),  # isolation window lower offset
            ("MS:1000829", ["value"]),  # isolation window upper offset
            ("MS:1000512", ["value"]),  # filter string
        ],
    )

    return data



"""
This module provides functionality for calculating the masses of peptide
fragments.
"""

from __future__ import absolute_import, division

from collections import Counter

from . import masses, ms_labels

from .losses import PEPTIDE_LOSSES, INTERNAL_LOSSES, AA_LOSSES, MOD_LOSSES


DELTA_C13 = masses.exact_mass({"C": [-1, 1]})


def _sequence_mass(pep_seq):
    return sum(
        masses.AMINO_ACIDS[letter] +
        sum(
            masses.MODIFICATIONS[letter, mod]
            for mod in mods
        )
        for letter, mods in pep_seq
    )


def _sequence_name(pep_seq):
    return "".join(
        letter.lower() if mods else letter.upper()
        for letter, mods in pep_seq
        if letter not in ["N-term", "C-term", "C=O"]
    )


def _internal_fragment_ions(
    pep_seq,
    c13_num=0,
    any_losses=None, aa_losses=None, mod_losses=None,
):
    """
    Calculate the m/z of all internal fragmenets of a peptide.

    Parameters
    ----------
    pep_seq : list of tuple of (str, list of str)
        A list of peptide letters and each residue's modification(s).
    c13_num : int, optional
    aa_losses : list of str, optional
        Potential neutral losses for each fragment (i.e. Water, amine, CO).
        List is composed of neutral loss names.
    aa_losses : dict of str, str, optional
    mod_losses : dict of tuple of (str, str), list of str
        Potential neutral losses for modified amino acids (i.e. pY-HPO_3).
        Dictionary should map (letter, modification) to a list of neutral
        loss names.

    Returns
    -------
    generator of str, float
        Ion names and their corresponding m/z's.
    """
    if any_losses is None:
        any_losses = INTERNAL_LOSSES

    if aa_losses is None:
        aa_losses = AA_LOSSES

    if mod_losses is None:
        mod_losses = MOD_LOSSES

    # Calculate the mass of the peptide cut at any two sites between the N-
    # and C-terminus
    for start in range(2, len(pep_seq) - 2):
        for end in range(start + 1, len(pep_seq) - 1):
            # Only add the mass of an N-terminus, cleavage will be between
            # C=O and N-H bond, adding a hydrogen to N-H
            fragment = (
                [("N-term", ())] +
                list(pep_seq[start:end]) +
                [("C=O", ())]
            )

            mass = _sequence_mass(fragment)
            name = _sequence_name(fragment)

            # Also calculate any potential neutral losses from this fragment
            for loss_name, loss_mass in _generate_losses(
                pep_seq=fragment,
                max_depth=1,
                c13_num=c13_num,
                any_losses=any_losses,
                aa_losses=aa_losses,
                mod_losses=mod_losses,
            ):
                yield name + loss_name, mass + loss_mass


def _get_frag_masses(pep_seq):
    return [
        _sequence_mass([pep_seq[index]])
        for index in range(len(pep_seq))
    ]


def _charged_m_zs(name, mass, max_charge):
    for charge in range(1, max_charge + 1):
        yield (
            (
                name +
                (
                    "^{{{:+}}}".format(charge)
                    if charge > 1 else
                    "^{+}"
                )
            ),
            (mass + charge * masses.PROTON) / charge,
        )


def _format_loss(name, count):
    if not name or count == 0:
        return ""

    return "{}{}{}".format(
        "+" if count > 0 else "-",
        "{} ".format(abs(count)) if abs(count) > 1 else "",
        name,
    )


def _generate_c13(c13_num):
    for c13 in range(0, c13_num + 1):
        c13_mass = c13 * DELTA_C13
        c13_name = _format_loss("^{13}C", c13)
        yield c13_name, c13_mass


def _generate_losses(
    pep_seq=None,
    max_depth=2,
    c13_num=0,
    any_losses=None, aa_losses=None, mod_losses=None,
):
    def _generate_loss_combos(
        seq=None,
        losses=None, max_depth=2,
    ):
        # TODO: Check if duplicates are generated here
        if losses is None:
            losses = Counter({"": 0})
            yield losses

        if not seq:
            return

        if max_depth < 1:
            yield losses
            return

        for loss in any_losses:
            new_losses = losses.copy()

            for ion in loss.split("-"):
                new_losses[ion] -= 1

            yield new_losses

            child_losses = _generate_loss_combos(
                seq, new_losses,
                max_depth=max_depth - 1,
            )

            for new_losses in child_losses:
                yield new_losses

        for aa, a_losses in aa_losses.items():
            for index, (letter, mods) in enumerate(seq):
                if aa != letter:
                    continue

                new_seq = seq[:index] + seq[index + 1:]

                for loss in a_losses:
                    new_losses = losses.copy()

                    for ion in loss.split("-"):
                        new_losses[ion] -= 1

                    yield new_losses

                    child_losses = _generate_loss_combos(
                        new_seq, new_losses,
                        max_depth=max_depth - 1,
                    )

                    for new_losses in child_losses:
                        yield new_losses

        for (aa, mod), m_losses in mod_losses.items():
            for index, (letter, mods) in enumerate(seq):
                if aa != letter or (mod and mod not in mods):
                    continue

                new_seq = seq[:index] + seq[index + 1:]

                for loss in m_losses:
                    new_losses = losses.copy()

                    for ion in loss.split("-"):
                        new_losses[ion] -= 1

                    yield new_losses

                    child_losses = _generate_loss_combos(
                        new_seq, new_losses,
                        max_depth=max_depth - 1,
                    )

                    for new_losses in child_losses:
                        yield new_losses

    for loss in _generate_loss_combos(
        seq=pep_seq,
        max_depth=max_depth,
    ):
        loss_mass = sum(
            masses.MASSES[name] * count
            for name, count in loss.items()
            if name
        )
        loss_name = "".join(
            _format_loss(name, count)
            for name, count in sorted(loss.items(), key=lambda x: x[0])
        )

        for c13_name, c13_mass in _generate_c13(c13_num):
            yield loss_name + c13_name, loss_mass + c13_mass


def _b_y_ions(
    pep_seq, frag_masses, fragment_max_charge,
    c13_num=0,
    any_losses=None, aa_losses=None, mod_losses=None,
):
    def _generate_ions(seq, mass, basename):
        # b/y ions +/- losses
        for loss_name, loss_mass in _generate_losses(
            pep_seq=seq,
            c13_num=c13_num,
            any_losses=any_losses,
            aa_losses=aa_losses,
            mod_losses=mod_losses,
        ):
            for name, mz in _charged_m_zs(
                basename + loss_name,
                mass + loss_mass,
                fragment_max_charge,
            ):
                yield name, mz

    base_ions = {}

    for index in range(2, len(pep_seq) - 1):
        # XXX: iTRAQ / TMT y-adducts?
        base_ions["a_{{{}}}".format(index - 1)] = (
            sum(frag_masses[:index]) - masses.MASSES["CO"],
            pep_seq[:index],
        )
        base_ions["b_{{{}}}".format(index - 1)] = (
            sum(frag_masses[:index]),
            pep_seq[:index],
        )

    for index in range(1, len(pep_seq) - 1):
        # y ion: 1 hydrogen added to NH group, one hydrogen on K/R
        base_ions["y_{{{}}}".format(len(pep_seq) - index - 1)] = (
            sum(frag_masses[index:]) + masses.PROTON,
            pep_seq[index:],
        )

    for ion_name, (ion_mass, seq) in base_ions.items():
        for name, mz in _generate_ions(
            seq, ion_mass, ion_name,
        ):
            yield name, mz


def _label_ions(pep_seq):
    label_mods = [
        mod
        for mod in pep_seq[0][1]
        if mod in ms_labels.LABEL_NAMES
    ]

    for mod in label_mods:
        for name, mz in zip(
            ms_labels.LABEL_NAMES[mod],
            ms_labels.LABEL_MASSES[mod],
        ):
            yield name, mz


def _parent_ions(
    pep_seq, frag_masses, parent_max_charge,
    c13_num=0,
    any_losses=None, aa_losses=None, mod_losses=None,
):
    parent_mass = sum(frag_masses) + masses.PROTON

    for loss_name, loss_mass in _generate_losses(
        pep_seq=pep_seq,
        c13_num=c13_num,
        any_losses=any_losses,
        aa_losses=aa_losses,
        mod_losses=mod_losses,
    ):
        for name, mz in _charged_m_zs(
            "MH" + loss_name,
            parent_mass + loss_mass,
            parent_max_charge,
        ):
            yield name, mz


def _py_ions(pep_seq, c13_num=0):
    if any(
        letter == "Y" and "Phospho" in mods
        for letter, mods in pep_seq
    ):
        name = "pY"
        mass = (
            masses.IMMONIUM_IONS["Y"] +
            masses.MODIFICATIONS["Y", "Phospho"]
        )

        for loss_name, loss_mass in _generate_losses(
            c13_num=c13_num,
        ):
            yield name + loss_name, mass + loss_mass


def fragment_ions(
    pep_seq, charge,
    parent_max_charge=None, fragment_max_charge=None,
    c13_num=0,
    any_losses=None, aa_losses=None, mod_losses=None,
):
    """
    Calculate the m/z of all ions generated by a peptide.

    Parameters
    ----------
    pep_seq : list of tuple of (str, list of str)
    charge : int
    parent_max_charge : int, optional
    fragment_max_charge : int, optional
    c13_num : int, optional
    any_losses : list of str, optional
        Potential neutral losses for each fragment (i.e. Water, amine, CO).
        List is composed of neutral loss names.
    aa_losses : dict of str, str, optional
    mod_losses : dict of tuple of (str, str), str, optional
        Potential neutral losses for modified amino acids (i.e. pY-HPO_3).
        Dictionary should map (letter, modification) to a list of neutral
        loss names.

    Returns
    -------
    dict of int, dict of str, float
        Dictionary mapping fragment position to a dictionary mapping ion names
        to ion m/z's.
    """
    assert "N-term" == pep_seq[0][0]
    assert "C-term" == pep_seq[-1][0]

    if parent_max_charge is None:
        parent_max_charge = charge

    if fragment_max_charge is None:
        # TODO: This correct?
        fragment_max_charge = parent_max_charge - 1

    if any_losses is None:
        any_losses = PEPTIDE_LOSSES

    if aa_losses is None:
        aa_losses = AA_LOSSES

    if mod_losses is None:
        mod_losses = MOD_LOSSES

    # First calculate the masses of each residue along the backbone
    frag_masses = _get_frag_masses(pep_seq)

    frag_ions = {}

    # Get b/y (and associated a/c/x/z) ions
    frag_ions.update(
        _b_y_ions(
            pep_seq, frag_masses, fragment_max_charge,
            c13_num=c13_num,
            any_losses=any_losses,
            aa_losses=aa_losses,
            mod_losses=mod_losses,
        )
    )

    # Get parent ions (i.e. MH^{+1})
    frag_ions.update(
        _parent_ions(
            pep_seq, frag_masses, parent_max_charge,
            c13_num=c13_num,
            any_losses=any_losses,
            aa_losses=aa_losses,
            mod_losses=mod_losses,
        )
    )

    # Get TMT / iTRAQ labels
    frag_ions.update(
        _label_ions(pep_seq)
    )

    # Get pY peak
    frag_ions.update(
        _py_ions(
            pep_seq,
            c13_num=c13_num,
        )
    )

    # Get internal fragments
    frag_ions.update(
        _internal_fragment_ions(
            pep_seq,
            c13_num=c13_num,
            aa_losses=aa_losses,
            mod_losses=mod_losses,
        )
    )

    return frag_ions



"""
Provides functionality for interacting with MASCOT data.
"""

from __future__ import absolute_import, division

import logging
import xml.etree.ElementTree as ET

from . import pep_query, regexes


MASCOT_NS = {
    "mascot":
        "http://www.matrixscience.com/xmlns/schema/mascot_search_results_2",
}

LOGGER = logging.getLogger("pycamv.mascot")


def _count_instances(pep_seq, letters):
    return sum(
        (["N-term"] + list(pep_seq) + ["C-term"]).count(letter)
        for letter in letters
    )


def _parse_letters(letters):
    """
    Turns a string of residue letters (i.e. "STY") into a list.

    Includes special provisions for N- and C-term modifications.

    Parameters
    ----------
    letters : str

    Returns
    -------
    tuple of str
    """
    if letters in ["N-term", "C-term"]:
        return (letters,)

    return tuple(letters)


def _get_fixed_var_mods(root):
    fixed_mods = [
        i.text
        for i in root.findall(
            "mascot:fixed_mods/mascot:modification/mascot:name",
            MASCOT_NS,
        )
    ]
    var_mods = [
        i.text
        for i in root.findall(
            "mascot:variable_mods/mascot:modification/mascot:name",
            MASCOT_NS,
        )
    ]
    return fixed_mods, var_mods


def _parse_mascot_2_4_1(root):
    filename = root.find("mascot:header/mascot:FILENAME", MASCOT_NS).text
    filename = filename.split(":", 1)[1].strip()
    fixed_mods, var_mods = _get_fixed_var_mods(root)

    scan_used = {}
    index = 0
    out = []

    for hit in root.findall("mascot:hits/mascot:hit", MASCOT_NS):
        accession = hit.find("mascot:protein", MASCOT_NS).get("accession")
        prot_desc = hit.find("mascot:protein/mascot:prot_desc", MASCOT_NS).text

        if "FGR" in prot_desc.upper() or "FGR" in accession.upper():
            LOGGER.info(accession, prot_desc)

        match = regexes.RE_MASCOT_DESCRIPTION.match(prot_desc)

        if match:
            prot_desc = match.group(1)

        for peptide in hit.findall("mascot:protein/mascot:peptide", MASCOT_NS):
            query = int(peptide.get("query"))
            rank = int(peptide.get("rank"))
            pep_score = float(peptide.find("mascot:pep_score", MASCOT_NS).text)
            exp_mz = float(peptide.find("mascot:pep_exp_mz", MASCOT_NS).text)
            exp_z = int(peptide.find("mascot:pep_exp_z", MASCOT_NS).text)
            pep_seq = peptide.find("mascot:pep_seq", MASCOT_NS).text

            pep_var_mods = peptide.find("mascot:pep_var_mod", MASCOT_NS).text

            if fixed_mods:
                pep_fixed_mods = [
                    regexes.RE_DYN_MODS.match(mod.strip()).group(3, 4)
                    for mod in fixed_mods
                ]
                pep_fixed_mods = [
                    (name, _parse_letters(letters))
                    for name, letters in pep_fixed_mods
                ]
                pep_fixed_mods = [
                    (_count_instances(pep_seq, letters), name, letters)
                    for name, letters in pep_fixed_mods
                ]
                pep_fixed_mods = [
                    (count, name, letters)
                    for count, name, letters in pep_fixed_mods
                    if count > 0
                ]
            else:
                pep_fixed_mods = []

            if pep_var_mods:
                # i.e. "2 Phospho (STY)""
                pep_var_mods = [
                    regexes.RE_DYN_MODS.match(mod.strip()).group(2, 3, 4)
                    for mod in pep_var_mods.split(";")
                ]
                pep_var_mods = [
                    (int(count) if count else 1, name, _parse_letters(letters))
                    for count, name, letters in pep_var_mods
                ]
            else:
                pep_var_mods = []

            scan = int(
                regexes.RE_SCAN_NUM.search(
                    peptide.find("mascot:pep_scan_title", MASCOT_NS).text
                ).group(2)
            )

            # scan_used
            if scan in scan_used:
                index_match, rank_match = scan_used[scan]

                if rank >= rank_match:
                    continue
                else:
                    del out[index_match]

            scan_used[scan] = (index, rank)
            out.append(
                pep_query.PeptideQuery(
                    accessions=[accession],
                    prot_descs=[prot_desc],
                    query=query,
                    filename=filename,
                    pep_score=pep_score,
                    pep_exp_mz=exp_mz,
                    pep_exp_z=exp_z,
                    pep_seq=pep_seq,
                    pep_var_mods=pep_var_mods,
                    pep_fixed_mods=pep_fixed_mods,
                    scan=scan,
                    quant_scan=scan,
                )
            )
            index += 1

    out = sorted(out, key=lambda x: x.scan)

    return fixed_mods, var_mods, out


def read_mascot_xml(xml_path):
    """
    Parse a MASCOT XML file.

    Parameters
    ----------
    xml_path : str
        Path to XML file.

    Returns
    -------
    fixed_mods : list of str
    var_mods : list of str
    out : list of :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    """
    tree = ET.parse(xml_path)
    root = tree.getroot()

    ms_version = root.find("mascot:header/mascot:MascotVer", MASCOT_NS).text
    ms_version = tuple(int(i) for i in ms_version.split("."))

    if ms_version >= (2, 4, 1):
        return _parse_mascot_2_4_1(root)
#     elif ms_version == "2.4.0":
#         return _parse_mascot_2_4_0(root)
#     elif ms_version == "2.3.02":
#         return _parse_mascot_2_3_02(root)
#     elif ms_version == "2.1.03":
#         return _parse_mascot_2_1_03(root)
    else:
        raise Exception("Unsupported Mascot Version: {}".format(ms_version))





MAT_CHOICES = {1: 'accept', 2: 'maybe', 3: 'reject'}


def load_mat_validation(mat_paths):
    # scipy breaks Ctrl-C interrupts when scipy for some distributions,
    # only import as necessary
    from scipy.io import loadmat

    validation_data = {}

    for mat_path in mat_paths:
        mat_data = loadmat(mat_path)

        for scan in mat_data['data'][0]:
            scan_num = scan['scan_number'][0][0][0][0]

            if 'fragments' not in scan.dtype.names:
                continue

            for ptm_combo in scan['fragments'][0][0][0]:
                seq = ptm_combo['seq'][0][0][0]
                choice = ptm_combo['status'][0][0][0][0]

                if choice == 0:
                    continue

                validation_data[scan_num, seq] = MAT_CHOICES[choice]

    return validation_data



"""
Provides functionality for exporting processed data to .camv files (JSON
format).
"""

from __future__ import absolute_import, division

import errno
import logging
import os
import sqlite3
from time import time

from . import sql

LOGGER = logging.getLogger("pycamv.export")

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def _extract_pep_seq(sequence):
    return "".join(
        letter
        for letter, _ in sequence
        if letter not in ["N-term", "C-term", "C=O"]
    )


def _extract_mods(sequence):
    return tuple(
        tuple(mods)
        for _, mods in sequence
    )


def _pep_mod_name(pep_seq, mods):
    return "".join(
        letter.lower() if mods else letter.upper()
        for letter, mods in zip(pep_seq, mods[1:-1])
    )


def export_to_sql(
    out_path, queue, scan_mapping, search_path, raw_paths,
    reprocess=False, total_num_seq=None,
):
    assert os.path.splitext(out_path)[1] in sql.DB_EXTS

    if not reprocess:
        try:
            os.remove(out_path)
        except FileNotFoundError as e:
            if type(e) == IOError and e.errno != errno.EEXIST:
                raise e

    db = sqlite3.connect(out_path, isolation_level="EXCLUSIVE")
    cursor = db.cursor()

    sql.create_tables(cursor)
    sql.run_migrations(cursor)

    sql.insert_path_data(cursor, search_path, raw_paths)

    total = time()

    # frag_map = defaultdict(list)
    index = 1
    while index <= total_num_seq:
        pep_query, seq, choice, peaks, precursor_win, label_win = queue.get()

        scan_query = scan_mapping[pep_query]

        LOGGER.debug(
            "Exporting: {} - {}{} {}".format(
                pep_query.scan,
                _pep_mod_name(
                    _extract_pep_seq(seq),
                    _extract_mods(seq),
                ),
                "- {}".format(choice) if choice else "",
                "({} / {})".format(index, total_num_seq)
                if total_num_seq else
                "(# {})".format(index),
            )
        )

        # Protein
        protein_ids = sql.insert_protein(cursor, pep_query)
        protein_set_id = sql.insert_protein_set(cursor, pep_query)

        # Peptide
        peptide_id = sql.insert_peptide(cursor, pep_query, protein_set_id)
        sql.insert_pep_prot(
            cursor,
            peptide_id, protein_ids,
            pep_query.pep_offsets,
        )

        # Modification data
        mod_state_id = sql.insert_mod_state(cursor, pep_query, peptide_id)
        ptm_id = sql.insert_ptm(cursor, pep_query, seq, mod_state_id)

        # Scan header data
        quant_mz_id = sql.insert_quant_mz(cursor, pep_query)
        file_id = sql.insert_file(cursor, pep_query)
        scan_id = sql.insert_scans(
            cursor, pep_query, scan_query,
            quant_mz_id,
            file_id,
            reprocessed=reprocess,
        )

        # Scan peak data
        sql.insert_peaks(
            cursor, peaks, scan_id,
        )
        sql.insert_precursor_peaks(
            cursor, scan_query, precursor_win, scan_id,
        )
        sql.insert_quant_peaks(
            cursor, pep_query, label_win, scan_id,
        )

        # PTM - Scan Mapping
        scan_ptm_id = sql.insert_scan_ptms(
            cursor, pep_query, scan_id, ptm_id,
            choice=choice,
        )
        sql.insert_fragments(cursor, peaks, scan_ptm_id)

        db.commit()
        LOGGER.debug(
            "done - avg: {:.3f} sec".format((time() - total) / (index + 1))
        )

        index += 1

    LOGGER.debug(
        "total: {:.3f} min ({:.3f} sec / peptide)"
        .format((time() - total) / 60, (time() - total) / (index + 1))
    )

    db.close()



"""
This module provides functionality for interacting with mass spec scan data.
"""

from __future__ import absolute_import, division

import logging
import os
import tempfile
import xml.etree.ElementTree as ET

from . import ms_labels, proteowizard, regexes


LOGGER = logging.getLogger("pycamv.scans")


class ScanQuery:
    """
    Attributes
    ----------
    scan : int
    isolation_mz : float or None
    window_offset : tuple of (int, int) or None
    precursor_scan : int or None
    collision_type : str or None
    c13_num : int
    basename : str or None
    """
    def __init__(
        self, scan,
        isolation_mz=None, window_offset=None, precursor_scan=None,
        collision_type=None, c13_num=0, basename=None,
    ):
        self.scan = scan
        self.precursor_scan = precursor_scan
        self.window_offset = window_offset
        self.isolation_mz = isolation_mz
        self.collision_type = collision_type
        self.c13_num = c13_num
        self.basename = basename


def _scanquery_from_spectrum(pep_query, spectrum):
    """
    Parameters
    ----------
    pep_query : :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    spectrum : :class:`pymzml.spec.Spectrum<spec.Spectrum>`

    Returns
    -------
    :class:`ScanQuery<pycamv.scans.ScanQuery>`
    """
    # prefix = {"mzml": "http://psi.hupo.org/ms/mzml"}

    scan = spectrum["id"]
    isolation_mz = spectrum["isolation window target m/z"]
    window_offset = (
        spectrum["isolation window lower offset"],
        spectrum["isolation window upper offset"],
    )
    collision_type = regexes.RE_COLLISION_TYPE.search(
        spectrum["filter string"]
    ).group(1).upper()

    ns = {"ns0": "http://psi.hupo.org/ms/mzml"}
    precursor = spectrum.xmlTreeIterFree.find(
        "ns0:precursorList/ns0:precursor",
        ns,
    ) or spectrum.xmlTreeIterFree.find(
        "precursorList/precursor",
    )

    if precursor is None:
        LOGGER.error(
            "Unable to find precursor scan info in scan {}".format(scan)
        )
        LOGGER.error(
            ET.tostring(
                spectrum.xmlTreeIterFree,
                encoding='utf8', method='xml',
            )
        )

    spectrum_ref = precursor.get("spectrumRef")

    precursor_scan = int(
        regexes.RE_PRECURSOR_SCAN.search(spectrum_ref).group(1)
    )

    return ScanQuery(
        scan,
        precursor_scan=precursor_scan,
        window_offset=window_offset,
        isolation_mz=isolation_mz,
        collision_type=collision_type,
        c13_num=_c13_num(pep_query, isolation_mz),
        basename=pep_query.basename,
    )


def _c13_num(pep_query, isolation_mz):
    """
    Counts the number of C13 atoms in a query, based on the mass-error between
    the expected and isolated m/z values.

    Parameters
    ----------
    pep_query : :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    isolation_mz : float

    Returns
    -------
    int
    """
    return int(
        round(
            pep_query.pep_exp_z *
            abs(pep_query.pep_exp_mz - isolation_mz)
        )
    )


def get_precursor_peak_window(
    scan_query,
    scan=None, ms_data=None, window_size=1
):
    """
    Get ion peaks around each peptide's precursor m/z range.

    Parameters
    ----------
    scan_query : :class:`ScanQuery<pycamv.scan_query.ScanQuery>`
    scan : :class:`pymzml.spec.Spectrum<spec.Spectrum>`, optional
    ms_data : :class:`pymzml.run.Reader<run.Reader>`, optional
    window_size : float, optional

    Returns
    -------
    list of list of tuple of (float, float)
    """
    assert scan is not None or ms_data is not None

    window = (
        scan_query.isolation_mz -
        (scan_query.window_offset[0] if scan_query.window_offset else 0) -
        window_size,
        scan_query.isolation_mz +
        (scan_query.window_offset[1] if scan_query.window_offset else 0) +
        window_size,
    )

    if scan is None:
        scan = ms_data[scan_query.basename][scan_query.precursor_scan]

    return [
        (mz, i)
        for mz, i in scan.centroidedPeaks
        if mz > window[0] and mz < window[1]
    ]


def get_label_peak_window(
    pep_query,
    scan=None, ms_data=None, window_size=1
):
    """
    Get ion peaks around each peptide's label m/z range.

    Parameters
    ----------
    pep_query : :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    scan : :class:`pymzml.spec.Spectrum<spec.Spectrum>`, optional
    ms_data : :class:`pymzml.run.Reader<run.Reader>`, optional
    window_size : float, optional

    Returns
    -------
    list of list of tuple of (float, float)
    """
    if scan is None and ms_data is None:
        return []

    label_mods = pep_query.get_label_mods

    if not label_mods:
        return []

    window = ms_labels.LABEL_MZ_WINDOW[label_mods[0]]
    window = (
        window[0] - window_size,
        window[1] + window_size
    )

    if scan is None:
        scan = ms_data[pep_query.basename][pep_query.quant_scan]

    return [
        (mz, i)
        for mz, i in scan.centroidedPeaks
        if mz > window[0] and mz < window[1]
    ]


def get_scan_data(raw_paths, pep_queries, out_dir=None):
    """
    Gets MS^2 and MS data for all scans in queries.

    Parameters
    ----------
    raw_path : list of str
    pep_queries : list of :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    out_dir : str, optional

    Returns
    -------
    scan_queries : list of :class:`ScanQuery<pycamv.scans.ScanQuery>`
    ms2_data : :class:`pymzml.run.Reader<run.Reader>`
    ms_data : :class:`pymzml.run.Reader<run.Reader>`
    """
    if out_dir is None:
        out_dir = tempfile.mkdtemp()

    # Collect MS^2 data
    ms_data = {}
    ms2_data = {}
    scan_queries = []

    for index, raw_path in enumerate(raw_paths):
        base_raw = os.path.basename(raw_path)

        ms2_scan_filter = sorted(
            set(
                scan
                for pep_query in pep_queries
                if pep_query.basename == os.path.basename(raw_path)
                for scan in (pep_query.scan, pep_query.quant_scan)
                if scan
            ),
        )

        if not ms2_scan_filter:
            continue

        LOGGER.info(
            "Converting MS^2 data for {} ({} / {})"
            .format(raw_path, index * 2 + 1, len(raw_paths) * 2)
        )

        ms2_data[base_raw] = proteowizard.raw_to_mzml(
            raw_path, os.path.join(out_dir, "ms2"),
            scans=ms2_scan_filter,
        )

        # Build a list of scan queries, including data about each scan
        scan_queries += [
            _scanquery_from_spectrum(
                pep_query,
                ms2_data[base_raw][pep_query.scan],
            )
            for pep_query in pep_queries
            if pep_query.basename == base_raw
        ]

        # Collect MS^1 data
        LOGGER.info(
            "Converting MS^1 data for {} ({} / {})"
            .format(raw_path, index * 2 + 2, len(raw_paths) * 2)
        )

        ms_scan_filter = sorted(
            set(
                scan_query.precursor_scan
                for scan_query in scan_queries
                if scan_query.basename == os.path.basename(raw_path)
            ),
        )

        ms_data[base_raw] = proteowizard.raw_to_mzml(
            raw_path, os.path.join(out_dir, "ms"),
            scans=ms_scan_filter,
        )

    return scan_queries, ms2_data, ms_data



"""
This module provides functionality for calculating molecular masses.
"""

# Built-ins
from __future__ import absolute_import, division

from collections import Iterable

# Core data analysis libraries


def exact_mass(atoms):
    """
    Parameters
    ----------
    atoms : dict of str, int or iterable
        Atoms in molecule and their count. If the dictionary's values are
        integers, use average isotope weights. If values are iterables, use
        exact count of each isotope, in order of increasing weight.

    Returns
    -------
    float
        Molecular weight of molecule.

    Examples
    --------
    >>> exact_mass({"H": 1})        # One hydrogen (Average isotope)
    1.007940721855
    >>> exact_mass({"H": [1]})      # One hydrogen (Exact isotope)
    1.007825
    >>> exact_mass({"H": [1, 1]})   # One hydrogen, one deuterium
    3.021927
    >>> exact_mass({"H": [0, 1]})   # One deuterium (Exact isotope)
    2.014102
    """
    # Atom name mapping to molecular weight and frequency of that isotope
    molecular_weights = {
        "H": [
            (1.007825, 99.9885),
            (2.014102,  0.0115),
        ],
        "C": [
            (12.000000, 98.93),
            (13.003355,  1.07)
        ],
        "N": [
            (14.003074, 99.632),
            (15.000109,  0.368),
        ],
        "O": [
            (15.994915, 99.757),
            (16.999131,  0.038),
            (17.999159,  0.205),
        ],
        "S": [
            (31.972072, 94.93),
            (32.971459,  0.76),
            (33.967868,  4.29),
            (35.967079,  0.02),
        ],
        "P": [
            (30.973763, 100),
        ],
    }
    return sum(
        sum(
            weight[0] * count
            for weight, count in zip(molecular_weights[atom], counts)
        )
        if isinstance(counts, Iterable) else
        sum(
            weight * counts * frequency / 100
            for weight, frequency in molecular_weights[atom]
        )
        for atom, counts in atoms.items()
    )


PROTON = exact_mass({"H": [1]})

C_TERM = exact_mass({"H": [1], "O": [1]})
N_TERM = exact_mass({"H": [1]})

# iTRAQ Masses not exact due to mass differences between isotopes
ITRAQ_4_PLEX = exact_mass({"C": [4, 3], "N": [1, 1], "O": [1], "H": [12]})
ITRAQ_8_PLEX = exact_mass({"C": [7, 7], "N": [3, 1], "O": [3], "H": [24]})
TMT_6_PLEX = exact_mass({"C": [8, 4], "N": [1, 1], "O": [2], "H": [20]})
TMT_10_PLEX = TMT_6_PLEX
TMT_11_PLEX = TMT_10_PLEX

LYSINE_HYDROGEN = exact_mass({"H": [1]})

ALANINE = exact_mass({"H": [5], "C": [3], "O": [1], "N": [1]})
ARGININE = exact_mass({"H": [12], "C": [6], "O": [1], "N": [4]})
ASPARAGINE = exact_mass({"H": [6], "C": [4], "O": [2], "N": [2]})
ASPARTATE = exact_mass({"H": [5], "C": [4], "O": [3], "N": [1]})
CYSTEINE = exact_mass({"H": [5], "C": [3], "O": [1], "N": [1], "S": [1]})
GLUTAMATE = exact_mass({"H": [7], "C": [5], "O": [3], "N": [1]})
GLUTAMINE = exact_mass({"H": [8], "C": [5], "O": [2], "N": [2]})
GLYCINE = exact_mass({"H": [3], "C": [2], "O": [1], "N": [1]})
HISTIDINE = exact_mass({"H": [7], "C": [6], "O": [1], "N": [3]})
ISOLEUCINE = exact_mass({"H": [11], "C": [6], "O": [1], "N": [1]})
LEUCINE = exact_mass({"H": [11], "C": [6], "O": [1], "N": [1]})
LYSINE = exact_mass({"H": [12], "C": [6], "O": [1], "N": [2]})
METHIONINE = exact_mass({"H": [9], "C": [5], "O": [1], "N": [1], "S": [1]})
PHENYLALANINE = exact_mass({"H": [9], "C": [9], "O": [1], "N": [1]})
PROLINE = exact_mass({"H": [7], "C": [5], "O": [1], "N": [1]})
SERINE = exact_mass({"H": [5], "C": [3], "O": [2], "N": [1]})
THREONINE = exact_mass({"H": [7], "C": [4], "O": [2], "N": [1]})
TYROSINE = exact_mass({"H": [9], "C": [9], "O": [2], "N": [1]})
TRYPTOPHAN = exact_mass({"H": [10], "C": [11], "O": [1], "N": [2]})
VALINE = exact_mass({"H": [9], "C": [5], "O": [1], "N": [1]})

AMINE = exact_mass({"N": [1], "H": [3]})
WATER = exact_mass({"H": [2], "O": [1]})
OXIDE = exact_mass({"O": [1]})
DIOXIDE = exact_mass({"O": [2]})
CARBON_MONOXIDE = exact_mass({"C": [1], "O": [1]})
CARBON_DIOXIDE = exact_mass({"C": [1], "O": [2]})
PHOSPHORIC_ACID = exact_mass({"H": [3], "P": [1], "O": [4]})
PHOSPHITE = exact_mass({"H": [1], "P": [1], "O": [3]})
SOCH4 = exact_mass({"S": [1], "O": [1], "C": [1], "H": [4]})
SO2CH4 = exact_mass({"S": [1], "O": [2], "C": [1], "H": [4]})
HYDROXYL = exact_mass({"H": [1], "O": [1]})
H2_PHOSPHATE = exact_mass({"H": [2], "O": [4], "P": [1]})
ACETYL = exact_mass({"H": [3], "C": [2], "O": [1]})
CARBAMIDOMETHYL = exact_mass({"H": [4], "C": [2], "O": [1], "N": [1]})

ACETYL_LYSINE = -PROTON + ACETYL
OXY_METHIONINE = OXIDE
DIOXY_METHIONINE = DIOXIDE
CARBAMIDOMETHYL_CYSTEINE = -PROTON + CARBAMIDOMETHYL
PHOSPHO_SERINE = -HYDROXYL + H2_PHOSPHATE
PHOSPHO_THREONINE = -HYDROXYL + H2_PHOSPHATE
PHOSPHO_TYROSINE = -HYDROXYL + H2_PHOSPHATE
PHOSPHO_TYROSINE_IMMONIUM = PHOSPHO_TYROSINE - CARBON_MONOXIDE + PROTON

SILAC_LYSINE_13C6 = exact_mass({"C": [-6, 6]})
SILAC_LYSINE_13C615N2 = exact_mass({"C": [-6, 6], "N": [-2, 2]})
SILAC_ARGININE_13C6 = exact_mass({"C": [-6, 6]})
SILAC_ARGININE_13C615N4 = exact_mass({"C": [-6, 6], "N": [-4, 4]})

# XXX: SILAC Tyrosine?
# XXX: SILAC Leucine?
# XXX: SILAC Acetyl-lysine? (Multiple modifications)

AMINO_ACIDS = {
    # Amino Acids
    "A": ALANINE,
    "R": ARGININE,
    "N": ASPARAGINE,
    "D": ASPARTATE,
    "C": CYSTEINE,
    "E": GLUTAMATE,
    "Q": GLUTAMINE,
    "G": GLYCINE,
    "H": HISTIDINE,
    "I": ISOLEUCINE,
    "L": LEUCINE,
    "K": LYSINE,
    "M": METHIONINE,
    "F": PHENYLALANINE,
    "P": PROLINE,
    "S": SERINE,
    "T": THREONINE,
    "Y": TYROSINE,
    "W": TRYPTOPHAN,
    "V": VALINE,

    # Unknown Amino Acid
    "X": 0,

    # N- and C-terminus
    "N-term": N_TERM,
    "C-term": C_TERM,
    "C=O": 0,
}

IMMONIUM_IONS = {
    residue: mass - CARBON_MONOXIDE + PROTON
    for residue, mass in AMINO_ACIDS.items()
}

MODIFICATIONS = {
    # Modified Amino Acids
    ("K", "Acetyl"): ACETYL_LYSINE,
    ("M", "Oxidation"): OXY_METHIONINE,
    ("M", "Dioxidation"): DIOXY_METHIONINE,
    ("C", "Carbamidomethyl"): CARBAMIDOMETHYL_CYSTEINE,
    ("S", "Phospho"): PHOSPHO_SERINE,
    ("T", "Phospho"): PHOSPHO_THREONINE,
    ("Y", "Phospho"): PHOSPHO_TYROSINE,

    # TMT / iTRAQ
    ("N-term", "iTRAQ4plex"): ITRAQ_4_PLEX - N_TERM,
    ("N-term", "iTRAQ8plex"): ITRAQ_8_PLEX - N_TERM,
    ("N-term", "TMT6plex"): TMT_6_PLEX - N_TERM,
    ("N-term", "TMT10plex"): TMT_10_PLEX - N_TERM,
    ("N-term", "TMT11plex"): TMT_11_PLEX - N_TERM,
    ("K", "iTRAQ4plex"): ITRAQ_4_PLEX,
    ("K", "iTRAQ8plex"): ITRAQ_8_PLEX,
    ("K", "TMT6plex"): TMT_6_PLEX,
    ("K", "TMT10plex"): TMT_10_PLEX,
    ("K", "TMT11plex"): TMT_11_PLEX,

    # SILAC
    ("K", "Lysine-13C6 (K-13C6)"): SILAC_LYSINE_13C6,
    ("K", "Lysine-13C615N2 (K-full)"): SILAC_LYSINE_13C615N2,
    ("R", "Arginine-13C6 (R-13C6)"): SILAC_ARGININE_13C6,
    ("R", "Arginine-13C615N2 (R-full)"): SILAC_ARGININE_13C615N4,
}

# Weird ProteomeDiscoverer Amino Acids
# Uniprot will annotate as X when amino acid is unknown
# and it seems PD tries to annotate them (based on the spectra?)
MODIFICATIONS.update(
    {
        ("X", "Mapping{}".format(aa)): AMINO_ACIDS[aa]
        for aa in AMINO_ACIDS
        if len(aa) == 1 and aa != "X"
    }
)

# Dictionary mapping amino acids and their modifications to weights
MASSES = {
    "Proton": PROTON,

    # Neural Losses
    "H_2O": WATER,
    "NH_3": AMINE,
    "CO": CARBON_MONOXIDE,
    "CO_2": CARBON_DIOXIDE,
    "SOCH_4": SOCH4,                # oxy-M
    "SO2CH4": SO2CH4,               # dioxy-M
    "H_3PO_4": PHOSPHORIC_ACID,     # pS/T
    "HPO_3": PHOSPHITE,
    "HPO_3-H_2O": PHOSPHORIC_ACID,  # pY
}



"""
Provides functionality for interacting with ProteomeDiscoverer data.
"""

from __future__ import absolute_import, division

from collections import Counter, defaultdict
import logging
import os
import sqlite3
import xml.etree.ElementTree as ET

from . import masses, regexes, pep_query


LOGGER = logging.getLogger("pycamv.discoverer")


def _parse_letters(letters):
    """
    Turns a string of residue letters (i.e. "STY") into a list.

    Includes special provisions for N- and C-term modifications.

    Parameters
    ----------
    letters : str

    Returns
    -------
    tuple of str
    """
    if letters in ["N-term", "C-term"]:
        return (letters,)

    return tuple(letters)


def _get_fixed_var_mods_pd21(conn):
    query = conn.cursor().execute(
        """
        SELECT
        Workflows.WorkflowXML
        FROM
        Workflows
        """
    )

    fixed_mods = []
    var_mods = []

    for xml, in query:
        root = ET.fromstring(xml)
        params = root.findall(
            "WorkflowTree/WorkflowNode/"
            "ProcessingNodeParameters/ProcessingNodeParameter"
        )
        for param in params:
            if not param.get("IsValueSet", False) or not param.text:
                continue

            if param.get("Name", "").startswith("DynMod_"):
                var_mods.append(param.text)
            elif param.get("Name", "").startswith("StaticMod_"):
                fixed_mods.append(param.text)

    return fixed_mods, var_mods


def _get_fixed_var_mods(conn):
    try:
        query = conn.cursor().execute(
            """
            SELECT
            ProcessingNodeParameters.ParameterName,
            ProcessingNodeParameters.ParameterValue
            FROM
            ProcessingNodeParameters
            """,
        )
    except sqlite3.OperationalError:
        return _get_fixed_var_mods_pd21(conn)

    fixed_mods = []
    var_mods = []

    for name, val in query:
        if name.startswith("StaticMod_"):
            fixed_mods.append(val)
        elif name.startswith("DynMod_"):
            var_mods.append(val)

    return fixed_mods, var_mods


def _find_mod(abbrev, letter, pot_mods):
    for pot_mod in pot_mods:
        if pot_mod[0] == abbrev and letter in pot_mod[1]:
            return pot_mod

    return None


def _count_mods(mod_list):
    return [
        (count, abbrev, _parse_letters(letters))
        for (abbrev, letters), count in Counter(mod_list).items()
    ]


def _get_pep_mods(conn, pep_id, pep_seq, var_mods, fixed_mods):
    aa_mods = conn.cursor().execute(
        """
        SELECT
        Peptides.SearchEngineRank,
        AminoAcidModifications.ModificationName,
        PeptidesAminoAcidModifications.Position
        FROM Peptides
        JOIN PeptidesAminoAcidModifications
        JOIN AminoAcidModifications
        ON PeptidesAminoAcidModifications.AminoAcidModificationID=
        AminoAcidModifications.AminoAcidModificationID
        WHERE
        Peptides.PeptideID=:pepID AND
        PeptidesAminoAcidModifications.PeptideID=:pepID
        """,
        {
            "pepID": pep_id,
        }
    )

    pep_var_mods = []
    pep_fixed_mods = []
    rank_pos = defaultdict(set)

    for rank, abbrev, pos in sorted(
        aa_mods,
        key=lambda x: 0 if x[1].startswith("Mapping") else 1,
    ):
        letter = pep_seq[pos]
        rank_pos[rank].add((pos, abbrev))

        mod = _find_mod(abbrev, letter, var_mods)

        if mod:
            pep_var_mods.append(mod)
            continue

        mod = _find_mod(abbrev, letter, fixed_mods)

        if mod:
            pep_fixed_mods.append(mod)
            continue

        # Weird ProteomeDiscoverer setting for proteins with unknown amino
        # acids. We can remap them to their correct amino acid:
        if (
            letter == "X" and
            (letter, abbrev) in masses.MODIFICATIONS and
            abbrev.startswith("Mapping")
        ):
            pep_seq = pep_seq[:pos] + abbrev[-1] + pep_seq[pos + 1:]
            continue

        raise Exception(
            "Unexpected modification: {} {}".format(letter, abbrev)
        )

    term_mods = conn.cursor().execute(
        """
        SELECT
        AminoAcidModifications.ModificationName,
        AminoAcidModifications.PositionType
        FROM PeptidesTerminalModifications
        JOIN AminoAcidModifications
        ON PeptidesTerminalModifications.TerminalModificationID=
        AminoAcidModifications.AminoAcidModificationID
        WHERE
        PeptidesTerminalModifications.PeptideID=:pepID
        """,
        {
            "pepID": pep_id,
        }
    )

    for abbrev, pos_type in term_mods:
        letter = "N-term" if pos_type == 1 else "C-term"
        mod = _find_mod(abbrev, letter, var_mods)

        if mod:
            pep_var_mods.append(mod)
            continue

        mod = _find_mod(abbrev, letter, fixed_mods)

        if mod:
            pep_fixed_mods.append(mod)
            continue

        raise Exception(
            "Unexpected modification: {} {}".format(letter, abbrev)
        )

    pep_var_mods = _count_mods(pep_var_mods)
    pep_fixed_mods = _count_mods(pep_fixed_mods)

    return pep_seq, pep_var_mods, pep_fixed_mods, rank_pos


def _get_prot_info(conn, peptide_id):
    proteins = conn.cursor().execute(
        """
        SELECT
        ProteinAnnotations.Description,
        Proteins.Sequence
        FROM PeptidesProteins
        JOIN ProteinAnnotations
        ON ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID
        JOIN Proteins
        ON Proteins.ProteinID=PeptidesProteins.ProteinID
        WHERE
        PeptidesProteins.PeptideID=:pepID
        """,
        {
            "pepID": peptide_id,
        }
    )

    (
        accessions,
        descriptions,
        uniprot_accessions,
        full_seqs,
    ) = [], [], [], []

    for full_prot_desc, sequence in proteins:
        uniprot_accession, accession, prot_desc = \
            regexes.RE_DISCOVERER_DESCRIPTION.match(
                full_prot_desc,
            ).group(1, 2, 3)

        if not accession:
            raise Exception(
                "Unable to find accession ID for {}".format(full_prot_desc)
            )

        accessions.append(accession)
        descriptions.append(prot_desc)
        uniprot_accessions.append(uniprot_accession)
        full_seqs.append(sequence)

    return (
        accessions,
        descriptions,
        uniprot_accessions,
        full_seqs,
    )


def get_quant_scan(conn, scan):
    query = conn.cursor().execute(
        """
        SELECT
        quant_scans.FirstScan
        FROM ReporterIonQuanResultsSearchSpectra mapping
        JOIN SpectrumHeaders quant_scans
        ON mapping.SpectrumID=quant_scans.SpectrumID
        JOIN SpectrumHeaders search_scans
        ON search_scans.SpectrumID=mapping.SearchSpectrumID
        WHERE search_scans.FirstScan=:scanId
        """,
        {
            "scanId": scan,
        }
    )
    for quant_scan, in query:
        return quant_scan


def _get_peptide_queries(conn, fixed_mods, var_mods):
    out = []
    # index = 0
    # scan_used = {}
    fixed_mods = [
        regexes.RE_DYN_MODS.match(i).group(3, 4)
        for i in fixed_mods
    ]
    var_mods = [
        regexes.RE_DYN_MODS.match(i).group(3, 4)
        for i in var_mods
    ]

    try:
        # ProteomeDiscoverer == 1.4
        query = conn.cursor().execute(
            """
            SELECT
            Peptides.PeptideID,
            Peptides.Sequence,
            PeptideScores.ScoreValue,
            SpectrumHeaders.FirstScan,
            Masspeaks.Mass,
            Masspeaks.Charge,
            SpectrumHeaders.Charge,
            FileInfos.FileName
            FROM Peptides
            JOIN PeptideScores
            ON PeptideScores.PeptideID=Peptides.PeptideID
            JOIN PeptidesProteins
            ON PeptidesProteins.PeptideID=Peptides.PeptideID
            JOIN ProteinAnnotations
            ON ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID
            JOIN SpectrumHeaders
            ON SpectrumHeaders.SpectrumID=Peptides.SpectrumID
            JOIN FileInfos
            ON FileInfos.FileID=MassPeaks.FileID
            JOIN Masspeaks
            ON Masspeaks.MassPeakID=SpectrumHeaders.MassPeakID
            WHERE
            Peptides.SearchEngineRank=1
            """
        )
    except sqlite3.OperationalError:
        # ProteomeDiscoverer == 2.1
        query = conn.cursor().execute(
            """
            SELECT
            Peptides.PeptideID,
            Peptides.Sequence,
            PeptideScores.ScoreValue,
            SpectrumHeaders.FirstScan,
            Masspeaks.Mass,
            Masspeaks.Charge,
            SpectrumHeaders.Charge,
            WorkflowInputFiles.FileName
            FROM Peptides
            JOIN PeptideScores
            ON PeptideScores.PeptideID=Peptides.PeptideID
            JOIN PeptidesProteins
            ON PeptidesProteins.PeptideID=Peptides.PeptideID
            JOIN ProteinAnnotations
            ON ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID
            JOIN SpectrumHeaders
            ON SpectrumHeaders.SpectrumID=Peptides.SpectrumID
            JOIN WorkflowInputFiles
            ON WorkflowInputFiles.FileID=MassPeaks.FileID
            JOIN Masspeaks
            ON Masspeaks.MassPeakID=SpectrumHeaders.MassPeakID
            WHERE
            Peptides.SearchEngineRank=1
            """
        )

    for (
        pep_id, pep_seq, pep_score,
        scan, exp_mz, mass_z, exp_z, filename,
    ) in query:
        if exp_z < 1:
            LOGGER.warning(
                "Charge: {}, {} for {} (Scan {})"
                .format(exp_z, mass_z, pep_seq, scan)
            )

        pep_seq, pep_var_mods, pep_fixed_mods, rank_pos = _get_pep_mods(
            conn, pep_id, pep_seq, var_mods, fixed_mods,
        )

        quant_scan = get_quant_scan(conn, scan)

        (
            accessions,
            descriptions,
            uniprot_accessions,
            full_seqs,
        ) = _get_prot_info(conn, pep_id)

        out.append(
            pep_query.PeptideQuery(
                accessions=accessions,
                prot_descs=descriptions,
                uniprot_accessions=uniprot_accessions,
                full_seqs=full_seqs,
                query=pep_id,
                filename=filename,
                pep_score=pep_score,
                pep_exp_mz=exp_mz,
                pep_exp_z=exp_z,
                pep_seq=pep_seq,
                pep_var_mods=pep_var_mods,
                pep_fixed_mods=pep_fixed_mods,
                scan=scan,
                rank_pos=rank_pos,
                quant_scan=quant_scan,
            )
        )

    out = sorted(out, key=lambda x: x.scan)

    return out


def read_discoverer_msf(msf_path):
    """
    Parse a ProteomeDiscoverer MSF file.

    Parameters
    ----------
    msf_path : str
        Path to MSF file.

    Returns
    -------
    fixed_mods : list of str
    var_mods : list of str
    out : list of :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    """
    LOGGER.info(
        "Loading ProteomeDiscoverer peptides from \"{}\"".format(
            os.path.basename(msf_path),
        )
    )

    with sqlite3.connect(msf_path) as conn:
        fixed_mods, var_mods = _get_fixed_var_mods(conn)
        out = _get_peptide_queries(conn, fixed_mods, var_mods)

    return fixed_mods, var_mods, out
